import sys
import os
from os.path import dirname, join, basename, exists
from typing import List, Tuple, Callable, Union
import itertools
import multiprocessing as mp
from functools import partial
import logging

import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm
from PIL import Image, ImageDraw
from openslide import OpenSlide

from ._functional import (
    get_thumbnail,
    get_downsamples,
    try_thresholds,
    resize
)
from .preprocess.functional import preprocess, tissue_mask
from ._czi_reader import OpenSlideCzi
from .helpers._utils import (
    remove_extension,
    remove_images,
    save_pickle
)

__all__ = [
    'Cutter',
    'TMACutter'
]

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class Cutter(object):
    """
    Cut tiles from histological images.

    This class detectct tissue on the slide and cuts tiles of desired width
    from the image.

    Args:
        slide_path (str): Path to the slide image. All formats that are 
            supported by openslide can be used.
        width (int): Tile width.
        overlap (float, optional): Overlap between neighbouring tiles. Defaults 
            to 0.0.
        threshold (int or float, optional): Threshold value for tissue 
            detection. Defaults to 1.1.

            If threshold is an integer between [1, 255]:
                This value will be used as an threshold for tissue detection. 
                Different thresholds can be easily searched with the
                Cutter.try_thresholds() function.
            
            If threshold is a float:
                In this case, Otsu's binarization is used and the found 
                threshold is multiplied by `threshold` as Otsu isn't optimal
                for histological images.

        downsample (int, optional): Downsample used for the thumbnail.
            When a lower downsample is used, the thumbnail-based background 
            detection is more accurate but slower. Good results are achieved 
            with downsample=16. Defaults to 16.
        max_background (float, optional): Maximum amount of background allowed 
            for a tile. Due to the thumbnail-based background detection, tiles 
            with higher background percentage may pass through but rarely the 
            other way around. Defaults to 0.999.
        create_thumbnail (bool, optional):  Create a thumbnail if downsample is 
            not available. Defaults to False.
        thumbnail_path (str, optional): Load a created thumbnail from a file.
            Defaults to None.

    Raises:
        IOError: slide_path not found.
        ValueError: downsample is not available and create_thumbnail=False.
        IOError: thumbnail_path not found.

    Example::

        import histoprep as hp
        cutter = hp.Cutter(slide_path='path/to/slide', width=512, overlap=0.2)
        metadata = cutter.save('/path/to/output_dir')

    """

    def __init__(
            self,
            slide_path: str,
            width: int,
            overlap: float = 0.0,
            threshold: Union[int, float] = 1.1,
            downsample: int = 16,
            max_background: float = 0.999,
            create_thumbnail: bool = False,
            thumbnail_path: str = None):
        super().__init__()
        # Define slide reader.
        if not exists(slide_path):
            raise IOError(f'{slide_path} not found.')
        if slide_path.endswith('czi'):
            logger.warning(
                "Support for czi-files is in alpha phase! If "
                "you run into errors, please submit an issue to "
                "https://github.com/jopo666/HistoPrep/issues"
            )
            self.reader = OpenSlideCzi(slide_path)
            self._czi = True
        else:
            self.reader = OpenSlide(slide_path)
            self._czi = False
        # Assing basic stuff that user can see/check.
        self.slide_path = slide_path
        self.slide_name = remove_extension(basename(slide_path))
        self.dimensions = self.reader.dimensions
        self.downsample = downsample
        self.width = width
        self.overlap = overlap
        self.threshold = threshold
        self.max_background = max_background
        self.all_coordinates = self._get_all_coordinates()
        # Filter coordinates.
        if thumbnail_path is not None:
            if not exists(thumbnail_path):
                raise IOError(f'{thumbnail_path} not found.')
            self.thumbnail = Image.open(thumbnail_path).convert('RGB')
        else:
            self.thumbnail = get_thumbnail(
                slide_path=self.slide_path,
                downsample=self.downsample,
                create_thumbnail=create_thumbnail
            )
        if self.thumbnail is None:
            # Downsample not available.
            raise ValueError(
                f'Thumbnail not available for downsample {self.downsample}. '
                'Please set create_thumbnail=True or select downsample from\n'
                f'{self._downsamples()}'
            )
        self.threshold, self._tissue_mask = tissue_mask(
            image=self.thumbnail,
            threshold=self.threshold,
            return_threshold=True
        )
        self.filtered_coordinates = self._filter_coordinates()
        # Annotate thumbnail
        self._annotate()

    def __repr__(self):
        return self.__class__.__name__ + '()'

    def __len__(self):
        return len(self.filtered_coordinates)

    def available_downsamples(self):
        """
        Returns available downsamples for the slide.
        """
        print(self._downsamples())

    def _downsamples(self):
        string = 'Downsample  Dimensions'
        if self._czi:
            d = {1: self.dimensions}
        else:
            d = get_downsamples(self.slide_path)
        for item, val in d.items():
            string += f'\n{str(item).ljust(12)}{val}'
        return string

    def summary(self):
        """Returns a summary of the cutting process."""
        print(self._summary())

    def _summary(self):
        return (
            f"{self.slide_name}"
            f"\n  Tile width: {self.width}"
            f"\n  Tile overlap: {self.overlap}"
            f"\n  Threshold: {self.threshold}"
            f"\n  Max background: {self.max_background}"
            f"\n  Thumbnail downsample: {self.downsample}"
            f"\n  Total number of tiles: {len(self.all_coordinates)}"
            f"\n  After background filtering: {len(self.filtered_coordinates)}"
        )

    def get_annotated_thumbnail(self,
                                max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the annotated thumbnail for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image 
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Annotated thumbnail.
        """
        return resize(self._annotated_thumbnail, max_pixels)

    def get_thumbnail(self, max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the thumbnail for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image 
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Thumbnail.
        """
        return resize(self.thumbnail, max_pixels)

    def get_tissue_mask(self, max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the tissue mask for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image 
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Tissue mask.
        """
        mask = self._tissue_mask
        # Flip for a nicer image
        mask = 1 - mask
        mask = mask/mask.max()*255
        mask = Image.fromarray(mask.astype(np.uint8))
        return resize(mask, max_pixels)

    def _prepare_directories(self, output_dir: str) -> None:
        out_dir = join(output_dir, self.slide_name)
        # Save paths.
        self._meta_path = join(out_dir, 'metadata.csv')
        self._thumb_path = join(out_dir, f'thumbnail_{self.downsample}.jpeg')
        self._annotated_path = join(out_dir, 'thumbnail_annotated.jpeg')
        self._param_path = join(out_dir, 'parameters.p')
        self._summary_path = join(out_dir, 'summary.txt')
        self._image_dir = join(out_dir, 'tiles')
        # Make dirs.
        os.makedirs(out_dir, exist_ok=True)
        os.makedirs(self._image_dir, exist_ok=True)

    def _annotate(self) -> None:
        # Draw tiles to the thumbnail.
        self._annotated_thumbnail = self.thumbnail.copy()
        annotated = ImageDraw.Draw(self._annotated_thumbnail)
        w = h = int(self.width/self.downsample)
        for (x, y), __ in self.filtered_coordinates:
            x_d = round(x/self.downsample)
            y_d = round(y/self.downsample)
            annotated.rectangle([x_d, y_d, x_d+w, y_d+h],
                                outline='red', width=4)

    def try_thresholds(
        self,
        thresholds: List[int] = [250, 240, 230,
                                 220, 200, 190, 180, 170, 160, 150, 140],
        max_pixels: int = 1_000_000
    ) -> Image.Image:
        """
        Try out different thresholds for tissue detection.

        The function prepares tissue masks with given thresholds and slaps them
        all together in one summary image.

        Args:
            thresholds (List[int], optional): Thresholds to try. Defaults to 
                [250, 240, 230, 220, 200, 190, 180, 170, 160, 150, 140].
            max_pixels (int, optional): Downsample the image until the image 
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: [description]
        """
        return try_thresholds(thumbnail=self.thumbnail, thresholds=thresholds)

    def save(
        self,
        output_dir: str,
        overwrite: bool = False,
        image_format: str = 'jpeg',
        quality: int = 95,
        custom_preprocess: Callable[[Image.Image], dict] = None
    ) -> pd.DataFrame:
        """
        Save tile images and metadata.

        The function saves all the detected tiles in the desired format. When
        the acutal image is loaded into memory, basic preprocessing metrics are
        computed and added to metadata for preprocessing.

        Args:
            output_dir (str): Parent directory for all output.
            overwrite (bool, optional): This will **remove** all saved images, 
                thumbnail and metadata and save images again.. Defaults to 
                False.
            image_format (str, optional): Format can be jpeg or png. Defaults 
                to 'jpeg'.
            quality (int, optional): For jpeg compression. Defaults to 95.
            custom_preprocess (Callable[[Image.Image], dict], optional): This is
                intended for users that want to define their own preprocessing 
                function. The function must take a Pillow image as an input and 
                return a dictionary of desired metrics. Defaults to None.

        Raises:
            ValueError: Invalid image format.

        Returns:
            pd.DataFrame: Metadata.
        """

        allowed_formats = ['jpeg', 'png']

        if image_format not in allowed_formats:
            raise ValueError(
                'Image format {} not allowed. Select from {}'.format(
                    image_format, allowed_formats
                ))
        self._prepare_directories(output_dir)
        # Check if slide has been cut before.
        if exists(self._meta_path) and not overwrite:
            logger.warning(f'Slide has already been cut!')
            return pd.read_csv(self._meta_path)
        elif exists(self._meta_path) and overwrite:
            # Remove all previous files.
            os.remove(self._annotated_path)
            os.remove(self._meta_path)
            remove_images(self._image_dir)
        # Save both thumbnails.
        self.thumbnail.save(self._thumb_path, quality=95)
        self._annotated_thumbnail.save(self._annotated_path, quality=95)
        # Save used parameters. NOTE: Can't remember where I would need these...
        # self._save_parameters()
        # Save text summary.
        with open(self._summary_path, "w") as f:
            f.write(self._summary())
        # Wrap the saving function so it can be parallized.
        func = partial(save_tile, **{
            'slide_path': self.slide_path,
            'slide_name': self.slide_name,
            'image_dir': self._image_dir,
            'width': self.width,
            'threshold': self.threshold,
            'image_format': image_format,
            'quality': quality,
            'custom_preprocess': custom_preprocess,
        })
        # Multiprocessing to speed things up!
        metadata = []
        with mp.Pool(processes=os.cpu_count()-1) as p:
            for result in tqdm(
                p.imap(func, self.filtered_coordinates),
                total=len(self.filtered_coordinates),
                desc=self.slide_name,
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                metadata.append(result)
        metadata = list(filter(None, metadata))
        if len(metadata) == 0:
            logger.warning(f'No tiles saved from slide {self.slide_path}!')
            return
        # Save metadata.
        self.metadata = pd.DataFrame(metadata)
        self.metadata.to_csv(self._meta_path, index=False)
        return self.metadata

    def _get_all_coordinates(self):
        """Return tile coordinates over the whole slide."""
        x = [0]
        y = [0]
        overlap_px = int(self.width*self.overlap)
        while x[-1] < self.dimensions[0]:
            x.append(x[-1] + self.width - overlap_px)
        x = x[:-1]
        while y[-1] < self.dimensions[1]:
            y.append(y[-1] + self.width - overlap_px)
        y = y[:-1]
        coordinates = list(itertools.product(x, y))
        return coordinates

    def _filter_coordinates(self):
        """Filter out coordinates with too much background."""
        filtered = []
        width_d = np.ceil(self.width/self.downsample).astype(int)
        for x, y in self.all_coordinates:
            y_d = int(y/self.downsample)
            x_d = int(x/self.downsample)
            mask = self._tissue_mask[y_d:y_d+width_d, x_d:x_d+width_d]
            if mask.size == 0:
                continue
            bg_perc = 1 - mask.sum()/mask.size
            if bg_perc < self.max_background:
                filtered.append(((x, y), bg_perc))
        return filtered


def save_tile(
        coords: Tuple[int, int, float],
        slide_path: str,
        slide_name: str,
        image_dir: str,
        width: int,
        threshold: int,
        image_format: str,
        quality: int,
        custom_preprocess: Callable[[Image.Image], dict] = None
) -> dict:
    """Saves a tile and returns metadata (parallizable)."""
    # Load slide as it can't be pickled...
    if slide_path.endswith('czi'):
        reader = OpenSlideCzi(slide_path)
    else:
        reader = OpenSlide(slide_path)
    (x, y), bg_estimate = coords
    # Prepare filename.
    filepath = join(image_dir, f'x-{x}_y-{y}')
    if image_format == 'png':
        filepath = filepath + '.png'
    else:
        filepath = filepath + '.jpeg'
    # Collect basic metadata.
    metadata = {
        'path': filepath,
        'slide_name': slide_name,
        'x': x,
        'y': y,
        'width': width,
        'background_estimate': bg_estimate
    }
    # Load image.
    try:
        image = reader.read_region((x, y), 0, (width, width)).convert('RGB')
    except:
        # Sometimes parts of the slide are corrupt or something...
        return
    # Update metadata with preprocessing metrics.
    metadata.update(preprocess(image=image, threshold=threshold))
    # Add custom metrics.
    if custom_preprocess is not None:
        metadata.update(custom_preprocess(image))
    # Save image.
    if not exists(filepath):
        image.save(filepath, quality=quality)
    return metadata



import os
from os.path import join, basename, dirname, exists
from typing import List
import logging
import time
import multiprocessing as mp
from functools import partial

import cv2
import pandas as pd
import numpy as np
from tqdm import tqdm


__all__ = [
    'combine_metadata',
    'update_paths',
    'check_tiles'
]


class TqdmHandler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)

    def emit(self, record):
        msg = self.format(record)
        tqdm.write(msg)


# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
# logger.addHandler(TqdmHandler())


def combine_metadata(
        parent_dir: str,
        csv_path: str = None,
        overwrite: bool = False,
        tma_spots: bool = False
) -> pd.DataFrame:
    """
    Combines all metadata in ``parent_dir`` into a single csv-file.

    Args:
        parent_dir (str): Directory with all the processed tiles.
        csv_path (str, optional): Path for the combined metadata.csv.
            Doesn't have to be defined if you just want to return the 
            pandas dataframe. Defaults to ``None``.
        overwrite (bool, optional): Whether to overwrite if csv_path 
            exists. Defaults to ``False``.
        tma_spots (bool, optional): Wheter to combine spot metadata.
            Defaults to ``False``.

    Raises:
        IOError: ``parent_dir`` does not exist.
        IOError: ``csv_path`` exist and ``overwrite=False``.

    Returns:
        pd.DataFrame: Combined metadata.
    """
    if not os.path.exists(parent_dir):
        raise IOError(f'{parent_dir} does not exist.')
    if csv_path is not None and os.path.exists(csv_path) and not overwrite:
        raise IOError(f'{csv_path} exists and overwrite=False.')
    # Collect metadata paths.
    directories = [x.path for x in os.scandir(parent_dir) if x.is_dir()]
    if tma_spots:
        meta_name = 'spot_metadata'
    else:
        meta_name = 'metadata'
    meta_paths = [join(d, f'{meta_name}.csv') for d in directories]
    # Collect dataframes
    dataframes = multiprocessing_loop(
        func=get_metadata,
        loop_this=meta_paths,
        desc='Combining metadata',
        processes=1 if len(meta_paths) < 20 else None,
    )
    dataframes = list(filter(lambda x: x is not None, dataframes))
    if len(dataframes) == 0:
        logger.warning(f'No metadata found at {parent_dir}!')
        return
    metadata = pd.concat(dataframes)
    if csv_path is not None:
        logger.warning(f'Saving combined metadata to {csv_path}.')
        metadata.to_csv(csv_path, index=False)
    return metadata


def get_metadata(metadata_path):
    """Safely load metadata."""
    if not exists(metadata_path):
        # There might be slides that haven't been finished.
        logger.warning(
            f'{metadata_path} path not found! This warning might arise '
            'if you are still cutting slides, some of the slides were '
            'broken or no tissue was found on the slide.'
        )
        return None
    elif os.path.getsize(metadata_path) < 6:
        # There might be empty files.
        logger.warning(f'{metadata_path} is an empty file.')
        return None
    else:
        return pd.read_csv(metadata_path)


def update_paths(parent_dir: str):
    """Rename all paths in metadata.

    This function can be used if you move around/rename the folder with 
    metadata etc. In this case all paths saved in metadata.csv will be
    wrong and have to be updated.

    Args:
        `parent_dir` (str): Directory to loop through

    Raises:
        IOError: ``parent_dir`` does not exists
    """
    if not os.path.exists(parent_dir):
        raise IOError(f'Path {parent_dir} does not exists!')
    # Collect data directories to update.
    update_paths = []
    for f in os.scandir(parent_dir):
        if f.is_dir() and exists(join(f.path, 'metadata.csv')):
            update_paths.append(f.path)
    if len(update_paths) == 0:
        logger.warning(f'No data directories found at {parent_dir}.')
        return
    multiprocessing_loop(
        func=update_meta,
        loop_this=update_paths,
        desc='Updating paths',
        processes=1 if len(update_paths) < 20 else None
    )


def update_meta(data_dir: str):
    """Update metadata paths."""
    all_paths = []
    for x in os.scandir(data_dir):
        if x.name.endswith('metadata.csv'):
            all_paths.append(x.path)
    for path in all_paths:
        # Load metadata.
        meta = pd.read_csv(path)

        if 'spot_metadata' in path:
            # Update spot metadata.
            spot_paths = meta.path[~meta.path.isna()]
            meta.loc[~meta.path.isna(), 'path'] = update_spot_paths(
                spot_paths, data_dir)
        elif 'spot_path' in meta.columns:
            # Update spot and tile metadata from a TMA array.
            spot_paths = meta.spot_path[~meta.spot_path.isna()]
            meta.loc[~meta.spot_path.isna(), 'spot_path'] = update_spot_paths(
                spot_paths, data_dir)
            tile_paths = meta.path[~meta.path.isna()]
            meta.loc[~meta.path.isna(), 'path'] = update_tile_paths(
                tile_paths, data_dir, TMA=True)
        else:
            # Update only tile paths.
            tile_paths = meta.path[~meta.path.isna()]
            meta.loc[~meta.path.isna(), 'path'] = update_tile_paths(
                tile_paths, data_dir, TMA=False)
        # Save updated metadata.
        meta.to_csv(path, index=False)


def update_tile_paths(paths: list, data_dir: str, TMA: bool) -> list:
    new_paths = []
    for old in paths:
        filename = basename(old)
        if TMA:
            spot_dir = basename(dirname(old))
            subdir = join('tiles', spot_dir)
        else:
            subdir = 'tiles'
        new = join(data_dir, subdir, filename)
        new_paths.append(new)
    return new_paths


def update_spot_paths(paths: list, data_dir: str) -> list:
    new_paths = []
    for old in paths:
        filename = basename(old)
        new = join(data_dir, 'spots', filename)
        new_paths.append(new)
    return new_paths


def check_tiles(parent_dir: str, overwrite: bool = False) -> pd.DataFrame:
    """Checks each saved tile in the parent_dir.

    Loads each tile in the metadata.path column and saves the image 
    dimensions, number of NaN pixels and any file corruptions to 
    metadata.csv.

    Args:
        parent_dir (str): Directory of processed slides.

    Raises:
        IOError: parent_dir does not exists.

    Returns:
        pd.DataFrame: Combined dataframe with additional info for each image.
    """
    # Define logger.
    if not os.path.exists(parent_dir):
        raise IOError(f'Path {parent_dir} does not exists!')
    # Collect metadata paths to check.
    meta_paths = []
    for f in os.scandir(parent_dir):
        path = join(f.path, 'metadata.csv')
        if f.is_dir() and exists(path):
            meta_paths.append(path)
    if len(meta_paths) == 0:
        logger.warning(f'No metadata dataframes found inside {parent_dir}!')
        return
    # Check if metadata has been processed.
    func = partial(load_meta, **{'overwrite': overwrite})
    dataframes = multiprocessing_loop(
        func=func,
        loop_this=meta_paths,
        desc='Loading metadata',
    )
    dataframes = [x for x in dataframes if x is not None]
    if len(dataframes) == 0:
        logger.warning(
            f'All {len(meta_paths)} metadata dataframes have been processed! '
            'Please set overwrite=True if you want to re-check images.'
        )
        return
    total = str(len(dataframes))
    combined = []
    for i, (meta_path, metadata) in enumerate(dataframes):
        # Define desc and load metadata.
        current = str(i+1).rjust(len(total))
        desc = f'[{current}/{total}] {basename(dirname(meta_path))}'
        paths = metadata.path.tolist()
        # Check images.
        results = multiprocessing_loop(
            func=check_image,
            loop_this=paths,
            desc=desc,
        )
        # Unpack results.
        shapes = []
        nan_counts = []
        corruptions = []
        file_exists = []
        for d in results:
            shapes.append(d['shape'])
            nan_counts.append(d['nan_count'])
            corruptions.append(d['corrupted'])
            file_exists.append(d['exists'])
        # Add to metadata and save!
        metadata['shape'] = shapes
        metadata['nan_count'] = nan_counts
        metadata['corrupted'] = corruptions
        metadata['exists'] = file_exists
        metadata.to_csv(meta_path, index=False)
        # Add to combined list.
        combined.append(metadata)
        # Log findings to user.
        nonexistent = sum(~metadata.exists)
        corrupted = metadata.corrupted.sum()
        unique_shapes = len(
            metadata['shape'][~metadata['shape'].isna()].unique()
        )
        if nonexistent > 0 or corrupted > 0 or unique_shapes > 1:
            logger.warning('')
            if nonexistent > 0:
                logger.warning(
                    f'  Found {sum(~metadata.exists)} tiles that do not exist.')
            if corrupted > 0:
                logger.warning(
                    f'  Found {metadata.corrupted.sum()} corrupted images.')
            if unique_shapes > 1:
                logger.warning(
                    f'  Found {unique_shapes} unique shapes for images.')
            logger.warning('-'*40)
    # Combine each metadata and return to user.
    if len(combined) > 0:
        combined = pd.concat(combined)
    else:
        combined = None
    return combined

def load_meta(path, overwrite):
        metadata = pd.read_csv(path)
        if not overwrite and 'corrupted' in metadata.columns:
            return None
        else:
            return path, metadata


def check_image(path):
    """Collect info on the image."""
    if not exists(path):
        return {
            'shape': None,
            'nan_count': None,
            'corrupted': None,
            'exists': False
        }
    with open(path, 'rb') as f:
        check_chars = f.read()[-2:]
    # Check corruption
    if check_chars != b'\xff\xd9':
        return {
            'shape': None,
            'nan_count': None,
            'corrupted': True,
            'exists': True,
        }
    else:
        # Load image.
        image = cv2.imread(path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return {
            'shape': image.shape,
            'nan_count': np.isnan(image).sum(),
            'corrupted': False,
            'exists': True,
        }


def multiprocessing_loop(func, loop_this, desc=None, processes=None):
    """Use for easy multiprocessing."""
    results = []
    if processes is None:
        processes = os.cpu_count() - 1
    with mp.Pool(processes=processes) as p:
        for result in tqdm(
            p.imap(func, loop_this),
            total=len(loop_this),
            desc=desc,
            bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
        ):
            results.append(result)
    return results



import os
import time
from typing import List
import pickle
import multiprocessing as mp

import numpy as np
import pandas as pd
from tqdm import tqdm

__all__ = [
    'load_pickle',
    'save_pickle'
]


def load_pickle(path: str):
    """
    Load pickle from path.

    Args:
        path (str): Path to the pickle file.

    Raises:
        IOError: Path does not exist.

    Returns:
        [any]: File saved in a pickle.
    """
    if not os.path.exists(path):
        raise IOError(f'{path} does not exist!')
    with open(path, "rb") as f:
        data = pickle.load(f)
    return data


def save_pickle(data, path: str):
    """
    Save data to pickle.

    Args:
        data (any): Data to be saved.
        path (str): Filepath.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def remove_extension(path: str) -> str:
    """Return filename with the extension removed."""
    if '.' in path:
        return '.'.join(path.split('.')[:-1])
    else:
        return path


def remove_images(image_dir: str) -> None:
    """
    Remove all images in the image folder

    Args:
        image_dir (str): Directory to be removed
    """
    paths = [x.path for x in os.scandir(image_dir)]
    if len(paths) > 0:
        with mp.Pool(processes=os.cpu_count()) as p:
            for __ in tqdm(
                p.imap(remove, paths),
                total=len(paths),
                desc='Removing images',
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                continue


def remove(path: str) -> None:
    if os.path.exists(path):
        os.remove(path)


def flatten(l):
    """Flatten list of lists."""
    if all(isinstance(x, list) for x in l):
        return [item for sublist in l for item in sublist]
    else:
        return l


def format_seconds(n: int) -> str:
    """Format seconds into pretty string format."""
    days = int(n // (24 * 3600))
    n = n % (24 * 3600)
    hours = int(n // 3600)
    n %= 3600
    minutes = int(n // 60)
    n %= 60
    seconds = int(n)
    if days > 0:
        strtime = f'{days}d{(hours)}h:{minutes}m:{seconds}s'
    elif hours > 0:
        strtime = f'{(hours)}h:{minutes}m:{seconds}s'
    else:
        strtime = f'{minutes}m:{seconds}s'
    return strtime



import os
from os.path import join, exists
from typing import List, Union
import logging

import numpy as np
import pandas as pd
from PIL import Image, ImageDraw
from shapely.geometry import Polygon, MultiPolygon

from ._functional import resize

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TileLabeler():
    """Add labels for tiles.

        Args:
            data_dir (str): Data directory of the slide (created 
                during cutting). 

        Raises:
            IOError: `data_dir`, `metadata.csv` or the thumbnail image
                doesn't exist.
            ValueError: More than one width found from the metadata.
        """

    def __init__(
        self,
        data_dir: str,
    ):
        super().__init__()
        self.data_dir = data_dir
        self._meta_path = join(data_dir, 'metadata.csv')
        self._thumb_path = None
        # Check paths.
        if not exists(data_dir):
            raise IOError(f'{self._meta_path} does not exist!')
        if not exists(self._meta_path):
            raise IOError(f'{self._meta_path} does not exist!')
        for f in os.scandir(data_dir):
            if 'thumbnail_' in f.name and not 'annot' in f.name:
                self._thumb_path = f.path
        if self._thumb_path is None:
            raise IOError(f'Thumbnail was not found!')
        # Get metadata.
        self.metadata = pd.read_csv(self._meta_path)
        # Get necessary parameters for plotting.
        self.width = self.metadata['width'].unique()
        if self.width.size == 1:
            self.width = int(self.width)
        else:
            raise ValueError(
                'More than one width found from the metadata '
                f'({self.width.tolist()}).'
            )
        self.downsample = int(self._thumb_path.split('_')[-1].split('.')[0])
        # Load thumbnail.
        self._thumbnail = Image.open(self._thumb_path)
        self._annotated_thumbnail = False

    def _drop_labels(self, prefix):
        cols = self.metadata.columns.tolist()
        drop = [x for x in cols if prefix in x]
        return self.metadata.drop(columns=drop)

    def get_thumbnail(self, max_pixels=1_000_000) -> Image.Image:
        return resize(self._thumbnail, max_pixels)

    def get_annotated_thumbnail(self, max_pixels=1_000_000) -> Image.Image:
        if not self._annotated_thumbnail:
            logger.warning("You haven't created any labels yet!")
        else:
            return resize(self._annotated_thumbnail, max_pixels)

    def _annotate(self, mask: Union[Polygon, MultiPolygon]):
        """Draw the shapely mask to the thumbnail."""
        # Init mask image.
        overlay = Image.new('RGBA', self._thumbnail.size, (255, 255, 255, 0))
        draw = ImageDraw.Draw(overlay)
        if isinstance(mask, Polygon):
            mask = [mask]
        for polygon in mask:
            coords = []
            for x, y in polygon.exterior.coords:
                x = int(x/self.downsample)
                y = int(y/self.downsample)
                coords.append((x, y))
            draw.polygon(coords, fill=(255, 0, 0, 180), outline="blue")
        self._annotated_thumbnail = Image.alpha_composite(
            self._thumbnail.convert('RGBA'),
            overlay.convert('RGBA')
        ).convert('RGB')

    def label_from_shapely(
        self,
        mask: Union[Polygon, MultiPolygon],
        prefix: str,
        threshold: float,
        overwrite: bool = False,
    ) -> None:
        """Add labels and intersection percentages to metadata from a 
        shapely mask.

        Args:
            mask (Union[Polygon, MultiPolygon]): Annotation mask.
            prefix (str): Name for the label.
            threshold (float): How much overlap with mask is required to
                turn the label into 1.
            overwrite (bool, optional): Wheter to overwrite labels if 
                they already exist in the metadata. Defaults to False.

        Raises:
            ValueError: Mask is not in Shapely format.
        """
        if not isinstance(mask, Polygon) and not isinstance(mask, MultiPolygon):
            raise ValueError('Excpected {} or {} not {}.'.format(
                Polygon, MultiPolygon, type(mask)
            ))
        if (
            any(f'{prefix}_label' in x for x in self.metadata.columns)
            and not overwrite
        ):
            logger.warning(
                'This dataset has already been labeled with '
                f'prefix={prefix}! To overwrite, set overwrite=True.'
            )
        # Drop previous labels if found.
        self.metadata = self._drop_labels(prefix)
        if len(mask.bounds) == 0:
            # "Draw" a thumbnail.
            self._annotated_thumbnail = self._thumbnail.copy()
            # Empty mask so all are 0.
            rows = []
            for x in range(self.metadata.shape[0]):
                rows.append({
                    f'{prefix}_perc': 0.0,
                    f'{prefix}_label': 0,
                })
        else:
            # Draw thumbnail.
            self._annotate(mask)
            # Collect labels.
            coords = np.vstack((
                self.metadata.x.to_numpy(),
                self.metadata.y.to_numpy(),
                self.metadata.width.to_numpy()
            )).T
            # Shapely uses (minx, miny, maxx, maxy).
            rows = []
            for coord in coords:
                x, y, width = coord
                tile = Polygon([
                    (x, y), (x, y+width),  # lower corners
                    (x+width, y+width), (x+width, y)  # upper corners
                ])
                percentage = tile.intersection(mask).area/tile.area
                rows.append({
                    f'{prefix}_perc': percentage,
                    f'{prefix}_label': int(percentage > threshold)
                })
        # Concatenate to metadata.
        self.metadata = pd.concat([self.metadata, pd.DataFrame(rows)], axis=1)
        # Save thumbnail.
        path = join(self.data_dir, f'{prefix}_mask.jpeg')
        self._annotated_thumbnail.save(path)
        # Save new metadata.
        self.metadata.to_csv(self._meta_path, index=False)

    def __repr__(self):
        return self.__class__.__name__ + '()'
